from draughts.convert import move_from_variant
from typing import List, Optional, Any, Union
import warnings

warnings.simplefilter("default")


class StandardMove:
    """Convert between different move types. Uses internal move representation (white always starts)."""
    def __init__(self, board: Any = None, board_move: Optional[List[List[int]]] = None, hub_move: Optional[str] = None,
                 hub_position_move: Optional[str] = None, pdn_move: Optional[str] = None,
                 pdn_position_move: Optional[str] = None, steps_move: Optional[List[int]] = None,
                 li_api_move: Optional[List[str]] = None, li_one_move: Optional[str] = None,
                 has_captures: Optional[bool] = None, possible_moves: Optional[List[List[List[int]]]] = None,
                 possible_captures: Optional[List[List[Optional[int]]]] = None, hub_to_pdn_pseudolegal: bool = False,
                 variant: Optional[str] = None, is_null: Optional[bool] = None) -> None:
        self.ambiguous: Optional[bool] = None
        self.captures = None
        self.has_captures = has_captures
        if (possible_moves is None or possible_captures is None) and board:
            self.possible_moves, self.possible_captures = board.legal_moves()
        else:
            self.possible_moves = possible_moves
            self.possible_captures = possible_captures
        self.hub_to_pdn_pseudolegal = hub_to_pdn_pseudolegal
        self.variant = variant
        # if not self.variant and board:
        #     self.variant = board.variant
        self.original_pdn_move = pdn_move
        self.is_null = is_null

        if self.is_null is None:
            self.is_null = (board_move == [[0, 0]] or hub_move == '0-0' or hub_position_move == '0000' or
                            pdn_move == '0-0' or pdn_position_move == '0000' or steps_move == [0, 0] or
                            li_api_move == ['0000'] or li_one_move == '0000')

        if ((board_move or hub_move or hub_position_move or pdn_move or pdn_position_move or steps_move or li_api_move
             or li_one_move) and not self.is_null):
            if self.possible_moves:
                self.board_move = self._to_board(board_move, hub_move, hub_position_move, pdn_move, pdn_position_move,
                                                 steps_move, li_api_move, li_one_move)
                self.captures = self.possible_captures[self.possible_moves.index(self.board_move)]
                self.captures = [] if self.captures[0] is None else self.captures
                self.has_captures = bool(self.captures)
                self._from_board(hub_move, hub_position_move, pdn_move, pdn_position_move, steps_move, li_api_move,
                                 li_one_move)
            else:
                self._no_board(board_move, hub_move, hub_position_move, pdn_move, pdn_position_move, steps_move,
                               li_api_move, li_one_move)
        elif self.is_null:
            self.board_move = [[0, 0]]
            self.hub_move = '0-0'
            self.hub_position_move = '0000'
            self.pdn_move = '0-0'
            self.pdn_position_move = '0000'
            self.steps_move = [0, 0]
            self.li_api_move = ['0000']
            self.li_one_move = '0000'

    def _make_len_2(self, move: Union[str, int]) -> str:
        """
        Add a 0 in the front of the square if it is only 1 digit.
        e.g. The move 5 will be turned to 05 but the move 23 will be left the same.
        """
        return f'0{move}' if len(str(move)) == 1 else str(move)

    def _sort_captures(self, captures: List[Optional[int]]) -> str:
        """
        Sort the captures from the smallest number to the highest.
        e.g. [10, 30, 19] will change to '101930'.
        This function exists because hub engines return the captures in alphabetical order
        (e.g. for the move 231201 scan returns 23x01x07x18 instead of 23x01x18x07)
        """
        if captures and captures[0] is None:
            captures = []
        return ''.join(list(sorted(map(self._make_len_2, captures))))

    def _to_board(self, board_move_given: Optional[List[List[int]]] = None, hub_move: Optional[str] = None,
                  hub_position_move: Optional[str] = None, pdn_move: Optional[str] = None,
                  pdn_position_move: Optional[str] = None, steps_move: Optional[List[int]] = None,
                  li_api_move: Optional[List[str]] = None, li_one_move: Optional[str] = None) -> List[List[int]]:
        """Convert the move to a board_move. Requires a Game() object to make the conversions."""

        board_move = [] if board_move_given is None else board_move_given
        # Hub related move

        if hub_move:
            # Hub move
            if "-" in hub_move:
                hub_position_move = "".join(list(map(self._make_len_2, hub_move.split("-"))))
            else:
                hub_position_move = "".join(list(map(self._make_len_2, hub_move.split("x"))))

        # PDN related move

        if pdn_move:
            # PDN move
            pdn_move = move_from_variant(pdn_move, variant=self.variant)

            if "-" in pdn_move:
                pdn_position_move = "".join(list(map(self._make_len_2, pdn_move.split("-"))))
            else:
                pdn_position_move = "".join(list(map(self._make_len_2, pdn_move.split("x"))))

        # Hub related move

        if hub_position_move:
            # Hub position move

            # Order the captures
            hub_position_move_to_use = hub_position_move[:4] + self._sort_captures(
                [int(hub_position_move[i:i + 2]) for i in range(4, len(hub_position_move), 2)])

            moves_li_board = {}
            for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(
                    possible_move[-1][1]) + self._sort_captures(possible_capture)
                moves_li_board[li_move] = possible_move
            board_move = moves_li_board[hub_position_move_to_use]

        # PDN related move

        elif pdn_position_move:
            # PDN position move
            moves_li_board = {}
            if len(pdn_position_move) == 4:
                self.ambiguous = False
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1])
                    moves_li_board[li_move] = possible_move
            else:
                self.ambiguous = True
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    steps = [possible_move[0][0]]
                    for move in possible_move:
                        steps.append(move[1])
                    li_move = "".join(list(map(self._make_len_2, steps)))
                    moves_li_board[li_move] = possible_move
            board_move = moves_li_board[pdn_position_move]

        # Board related moves

        elif steps_move:
            # steps_move
            board_move = []
            for index in range(1, len(steps_move)):
                board_move.append([steps_move[index - 1], steps_move[index]])

        elif li_api_move:
            # li_api_move
            board_move = []
            for move in li_api_move:
                board_move.append([int(move[:2]), int(move[2:])])

        elif li_one_move:
            # li_one_move
            steps = [int(li_one_move[i:i + 2]) for i in range(0, len(li_one_move), 2)]
            board_move = []
            for index in range(1, len(steps)):
                board_move.append([steps[index - 1], steps[index]])

        return board_move

    def _from_board(self, hub_move: Optional[str] = None, hub_position_move: Optional[str] = None,
                    pdn_move: Optional[str] = None, pdn_position_move: Optional[str] = None,
                    steps_move: Optional[List[int]] = None, li_api_move: Optional[List[str]] = None,
                    li_one_move: Optional[str] = None) -> None:
        """Convert the move to all other move types. Requires a Game() object to make the conversions."""

        hub_move = "" if hub_move is None else hub_move
        hub_position_move = "" if hub_position_move is None else hub_position_move
        pdn_move = "" if pdn_move is None else pdn_move
        pdn_position_move = "" if pdn_position_move is None else pdn_position_move
        steps_move = [] if steps_move is None else steps_move
        li_api_move = [] if li_api_move is None else li_api_move
        li_one_move = "" if li_one_move is None else li_one_move

        # Board related move

        if not steps_move:
            # steps_move
            steps = [self.board_move[0][0]]
            for move in self.board_move:
                steps.append(move[1])
            steps_move = steps

        # Hub related moves

        if not hub_position_move:
            # Hub position move
            hub_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])
            if self.captures:
                sorted_captures = self._sort_captures(self.captures)
                sorted_captures_list = [sorted_captures[i:i + 2] for i in range(0, len(sorted_captures), 2)]
                hub_position_move += "".join(list(map(self._make_len_2, sorted_captures_list)))

        if not hub_move:
            # Hub move
            positions = [hub_position_move[i:i + 2] for i in range(0, len(hub_position_move), 2)]
            separator = "x" if self.captures else "-"
            hub_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        # PDN related moves

        if not pdn_position_move:
            # starts_endings contains the start position and the end position of every possible move.
            # It is used to detect ambiguous PDN moves, so we know when we have to expand the PDN move.
            # e.g. Use 3x14x23 instead of 3x23.

            # PDN position move

            start_end = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])
            same_start_ends = 0
            moves_with_same_start_end_and_captures = []
            for possible_move, possible_captures in zip(self.possible_moves, self.possible_captures):
                possible_start_end = self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1])
                if possible_start_end == start_end:
                    same_start_ends += 1
                    possible_steps_move = [possible_move[0][0]]
                    for semi_move in possible_move:
                        possible_steps_move.append(semi_move[1])
                    moves_with_same_start_end_and_captures.append(possible_steps_move)
            self.ambiguous = same_start_ends >= 2
            if self.ambiguous:
                # From FMJD (https://pdn.fmjd.org/grammar.html#pdn-3-0-grammar):
                # Disambiguated capture sequences have to specify a complete sequence of intermediate squares
                # along the path of the capture. If there is a change in direction, an intermediate square is
                # the square where a turn in direction was made. If there was not a change in direction,
                # the intermediate square is the square immediately behind a captured piece.
                #
                # This means that if we continue going in the same direction,
                # the intermediate square is the square immediately behind the captured piece.
                # pydraughts supports providing a square not immediately behind the captured piece, but when we
                # convert it to a PDN move, we have to change it to the square immediately behind the captured piece.
                correct_start_ends = []
                for index in range(1, len(steps_move) - 1):
                    closest_distance = 1000

                    for steps_move in moves_with_same_start_end_and_captures:
                        distance = abs(steps_move[index - 1] - steps_move[index])

                        # New closest square
                        if distance < closest_distance:
                            closest_distance = distance
                            correct_start_ends = [steps_move]
                        # The same intermediate square.
                        # This means that the difference is in a later intermediate square.
                        elif distance == closest_distance:
                            correct_start_ends.append(steps_move)
                        moves_with_same_start_end_and_captures = correct_start_ends

                correct_move = correct_start_ends[0]

                pdn_position_move = "".join(list(map(self._make_len_2, correct_move)))
            else:
                pdn_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])

        if not pdn_move:
            # PDN move
            positions = [pdn_position_move[i:i + 2] for i in range(0, len(pdn_position_move), 2)]
            separator = "x" if self.captures else "-"
            pdn_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        # Board related moves

        if not li_api_move:
            # li_api_move
            li_api_move = []
            for move in self.board_move:
                li_api_move.append(self._make_len_2(move[0]) + self._make_len_2(move[1]))

        if not li_one_move:
            # li_one_move
            li_one_move = "".join(list(map(self._make_len_2, steps_move)))

        self.hub_move = hub_move
        self.hub_position_move = hub_position_move
        self.pdn_move = pdn_move
        self.pdn_position_move = pdn_position_move
        self.steps_move = steps_move
        self.li_api_move = li_api_move
        self.li_one_move = li_one_move

    def _no_board(self, board_move_given: Optional[List[List[int]]] = None, hub_move: Optional[str] = None,
                  hub_position_move: Optional[str] = None, pdn_move: Optional[str] = None,
                  pdn_position_move: Optional[str] = None, steps_move: Optional[List[int]] = None,
                  li_api_move: Optional[List[str]] = None, li_one_move: Optional[str] = None) -> None:
        """Makes as many conversions as possible if the board was not given."""

        board_move = [] if board_move_given is None else board_move_given
        hub_move = "" if hub_move is None else hub_move
        hub_position_move = "" if hub_position_move is None else hub_position_move
        pdn_move = "" if pdn_move is None else pdn_move
        pdn_position_move = "" if pdn_position_move is None else pdn_position_move
        steps_move = [] if steps_move is None else steps_move
        li_api_move = [] if li_api_move is None else li_api_move
        li_one_move = "" if li_one_move is None else li_one_move

        # Board related moves

        if steps_move:
            # steps_move
            board_move = []
            for index in range(1, len(steps_move)):
                board_move.append([steps_move[index - 1], steps_move[index]])

        elif li_api_move:
            # li_api_move
            board_move = []
            for move in li_api_move:
                board_move.append([int(move[:2]), int(move[2:])])

        elif li_one_move:
            # li_one_move
            steps = [int(li_one_move[i:i + 2]) for i in range(0, len(li_one_move), 2)]
            board_move = []
            for index in range(1, len(steps)):
                board_move.append([steps[index - 1], steps[index]])

        if board_move:
            # steps_move
            steps = [board_move[0][0]]
            for board_move_part in board_move:
                steps.append(board_move_part[1])
            steps_move = steps

            # li_api_move
            li_api_move = []
            for board_move_part in board_move:
                move_part = self._make_len_2(board_move_part[0]) + self._make_len_2(board_move_part[1])
                li_api_move.append(move_part)

            # li_one_move
            li_one_move = "".join(list(map(self._make_len_2, steps_move)))

        # Hub related moves

        if hub_move:
            # Hub move
            if "-" in hub_move:
                self.has_captures = False
                hub_position_move = "".join(list(map(self._make_len_2, hub_move.split("-"))))
            else:
                self.has_captures = True
                hub_position_move = "".join(list(map(self._make_len_2, hub_move.split("x"))))
        elif hub_position_move:
            # Hub position move
            positions = [hub_position_move[i:i + 2] for i in range(0, len(hub_position_move), 2)]
            separator = "x" if self.has_captures else "-"
            hub_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if self.hub_to_pdn_pseudolegal:
            # Pseudo-legal PDN move
            pdn_position_move = hub_position_move[:4]
            separator = "x" if self.has_captures else "-"
            pdn_move = pdn_position_move[:2] + separator + pdn_position_move[2:]

        # PDN related moves

        if pdn_move:
            # PDN move
            pdn_move = move_from_variant(pdn_move, variant=self.variant)

            if "-" in pdn_move:
                self.has_captures = False
                pdn_position_move = "".join(list(map(self._make_len_2, pdn_move.split("-"))))
            else:
                self.has_captures = True
                pdn_position_move = "".join(list(map(self._make_len_2, pdn_move.split("x"))))
        elif pdn_position_move:
            # PDN position move
            positions = [pdn_position_move[i:i + 2] for i in range(0, len(pdn_position_move), 2)]
            separator = "x" if self.has_captures else "-"
            pdn_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if pdn_position_move:
            # Get if the move is ambiguous.
            self.ambiguous = not (len(pdn_position_move) == 4)

        self.board_move = board_move
        self.hub_move = hub_move
        self.hub_position_move = hub_position_move
        self.pdn_move = pdn_move
        self.pdn_position_move = pdn_position_move
        self.steps_move = steps_move
        self.li_api_move = li_api_move
        self.li_one_move = li_one_move



# This file is an adaptation of DXC100_draughts_client (https://github.com/akalverboer/DXC100_draughts_client) by akalverboer.

from __future__ import annotations
import threading
import logging
from draughts.engines.dxp_communication.dxp_classes import DamExchange, MySocket, GameStatus, DXP_WHITE, DXP_BLACK
from draughts import Move, WHITE
from typing import Optional

logger = logging.getLogger("pydraughts")


class Receiver:
    def __init__(self, sender: Sender) -> None:
        self.sender = sender
        self.accepted: Optional[bool] = None
        self.gameend_sent = False
        self.last_move: Optional[Move] = None
        self.last_move_changed = False
        self.listening = False
        self.receive_thread_started = False
        self.backreq_accepted: Optional[bool] = None
        self.takeback_in_progress = False
        self.receive_thread = threading.Thread(target=self.receive)

    def start(self) -> None:
        self.receive_thread_started = True
        self.receive_thread.start()

    def close(self) -> None:
        if self.receive_thread_started:
            self.receive_thread.join()

    def receive(self) -> None:
        logger.debug("DXP Client starts listening")
        self.listening = True
        while True:
            try:
                message = self.sender.socket.receive()  # wait for message
            except Exception as err:
                logger.debug(f"Error {err}")
                self.sender.socket.close()
                break

            message = message[0:127]  # DXP max length
            dxp_data = self.sender.dxp.parse(message)
            if dxp_data["type"] == "C":
                logger.debug(f"rcv CHAT: {message}")
                logger.debug(f"\nChat message: {dxp_data['text']}")

            elif dxp_data["type"] == "A":
                logger.debug(f"rcv GAMEACC: {message}")
                if dxp_data["accCode"] == "0":
                    self.sender.current.engine_color = self.sender.current.engine_color  # as requested
                    self.sender.current.engineName = dxp_data["engineName"]
                    logger.debug(f"\nGame request accepted by {dxp_data['engineName']}")
                    self.accepted = True
                    self.gameend_sent = False
                    self.sender.current.started = True
                else:
                    logger.debug(f"\nGame request NOT accepted by {dxp_data['engineName']} Reason: {dxp_data['accCode']}")
                    self.accepted = False
                    self.sender.current.started = False

            elif dxp_data["type"] == "E":
                logger.debug(f"rcv GAMEEND: {message}")
                logger.debug(f"\nRequest end of game accepted. Reason: {dxp_data['reason']} Stop: {dxp_data['stop']}")
                # Confirm game end by sending message back (if not sent by me)
                if self.sender.current.started and not self.gameend_sent:
                    self.sender.current.result = dxp_data["reason"]
                    msg = self.sender.dxp.msg_gameend(str(dxp_data["reason"]))  # `str` is there only for mypy.
                    self.sender.socket.send(msg)
                    logger.debug(f"snd GAMEEND: {msg}")
                    self.gameend_sent = True
                    self.sender.current.started = False

            elif dxp_data["type"] == "M":
                logger.debug(f"rcv MOVE: {message}")
                steps = [dxp_data['from'], dxp_data['to']]
                nsteps = list(map(int, steps))
                ntakes_int = list(map(int, dxp_data['captures']))
                ntakes = list(map(lambda pos: str(pos).zfill(2), sorted(ntakes_int)))
                logger.debug(f"FEN: {self.sender.current.pos.fen}, Steps: {nsteps}, Takes: {ntakes}")
                correct_move = None
                while True:
                    if not self.takeback_in_progress:
                        break
                for move in self.sender.current.pos.legal_moves():
                    if move.hub_position_move == f"{str(nsteps[0]).zfill(2)}{str(nsteps[-1]).zfill(2)}{''.join(ntakes)}":
                        correct_move = move

                if correct_move is not None:
                    self.sender.current.pos.push(correct_move)
                    self.last_move = correct_move
                    logger.debug(f"Move received: {correct_move.steps_move}")
                    self.last_move_changed = True
                else:
                    move_history = list(map(lambda old_move: old_move.steps_move, self.sender.current.pos.move_stack))
                    logger.debug(f"Error: received move is illegal [{message}]\nMove history: {move_history}")

            elif dxp_data["type"] == "B":
                # For the time being do not confirm request from server: send message back.
                logger.debug(f"rcv BACKREQ: {message}")
                acc_code = "1"  # 0: BACK YES; 1: BACK NO; 2: CONTINUE
                msg = self.sender.dxp.msg_backacc(acc_code)
                self.sender.socket.send(msg)
                logger.debug(f"snd BACKACC: {msg}")

            elif dxp_data["type"] == "K":
                # Answer to my request to move back
                logger.debug(f"rcv BACKACC: {message}")
                logger.debug(f"rcv BACKACC: {message}")  # TEST
                acc_code = dxp_data['accCode']
                if acc_code == "0":
                    # Actions to go back in history as specified in my request
                    self.backreq_accepted = True
                elif acc_code == "1":
                    logger.debug("Engine doesn't support going back.")
                    self.backreq_accepted = False
                else:
                    logger.debug("Engine wants to continue the game.")
                    self.backreq_accepted = False

            else:
                logger.debug(f"rcv UNKNOWN: {message}")
                logger.debug(f"\nrcv Unknown message: {message}")

        self.listening = False


class Sender:
    def __init__(self) -> None:
        self.dxp = DamExchange()
        self.socket = MySocket()
        self.current = GameStatus("W")
        self.receiver = Receiver(self)

    def setup(self, fen: str, variant: str) -> None:
        logger.debug(f"FEN: {fen.strip()}, Variant: {variant}")
        engine_color = DXP_BLACK if fen[0] == 'B' else DXP_WHITE
        self.current = GameStatus(fen=fen, engine_color=engine_color, variant=variant)

    def send_move(self, move: Move) -> None:
        logger.debug(f"FEN: {self.current.pos.fen}, Steps: {move.steps_move}, Captures: {move.captures}")
        time_spent = 0
        self.receiver.last_move_changed = False
        self.current.pos.push(move)
        msg = self.dxp.msg_move(move.steps_move, move.captures, time_spent)
        try:
            self.socket.send(msg)
            logger.debug(f"snd MOVE: {msg}")
        except Exception as err:
            logger.debug(f"Error sending move: {err}")
            return

    def connect(self, host: str, port: int) -> None:
        logger.debug(f"Host: {host}, Port: {port}")
        try:
            self.socket.open()
            self.socket.connect(host, int(port))  # with timeout
            logger.debug(f"Successfully connected to remote host {host}, port {port}")
        except Exception as err:
            self.socket.sock = None
            logger.debug(f"Error trying to connect: {err}")
            return
        self.receiver.start()

    def disconnect(self) -> None:
        logger.debug("Attempting to disconnect.")
        try:
            self.socket.close()
            logger.debug("Successfully closed socket.")
        except Exception as err:
            self.socket.sock = None
            logger.debug(f"Error trying to close socket: {err}")
        self.receiver.close()

    def chat(self, text: str) -> None:
        text = text.strip()  # trim whitespace
        logger.debug(f"Text: {text}")
        msg = self.dxp.msg_chat(text)
        try:
            self.socket.send(msg)
            logger.debug(f"snd CHAT: {msg}")
        except Exception as err:
            logger.debug(f"Error sending chat message: {err}")

    def gamereq(self, engine_color: str, time: int, max_moves: int) -> None:
        logger.debug(f"Engine color: {engine_color}, Time: {time}, Max moves: {max_moves}")
        dxp_color = DXP_WHITE if engine_color.upper().startswith('W') else DXP_BLACK
        msg = self.dxp.msg_gamereq(dxp_color, time, max_moves, self.current.pos, self.current.get_color())
        try:
            self.socket.send(msg)
            logger.debug(f"snd GAMEREQ: {msg}")
        except Exception as err:
            logger.debug(f"Error sending game request: {err}")

    def gameend(self) -> None:
        logger.debug("Attempting to send a gameend.")
        reason = "0"
        msg = self.dxp.msg_gameend(reason)
        self.receiver.gameend_sent = True
        try:
            self.socket.send(msg)
            logger.debug(f"snd GAMEEND: {msg}")
        except Exception as err:
            logger.debug(f"Error sending gameend message: {err}")

    def backreq(self, move: int, color: int) -> None:
        logger.debug(f"Request to return to {move}th move with {'WHITE' if color == WHITE else 'BLACK'} to move.")
        msg = self.dxp.msg_backreq(move, DXP_WHITE if color == WHITE else DXP_BLACK)
        self.receiver.last_move_changed = False
        self.receiver.takeback_in_progress = True
        try:
            self.socket.send(msg)
            logger.debug(f"snd BACKREQ: {msg}")
        except Exception as err:
            logger.debug(f"Error sending backreq request: {err}")



from __future__ import annotations
from draughts.core.game import Game, _convert_variant_names
from draughts.convert import (fen_from_variant, fen_to_variant, move_from_variant, move_to_variant,
                              _number_to_algebraic, _algebraic_to_number)
from draughts.core.move import StandardMove
import pickle
from typing import Optional, Any, List, Tuple

WHITE = 2
BLACK = 1


class Move(StandardMove):
    """Convert between different move types. Accounts for the differences between variants."""
    def __init__(self, board: Any = None, board_move: Optional[List[List[int]]] = None, hub_move: Optional[str] = None,
                 hub_position_move: Optional[str] = None, pdn_move: Optional[str] = None,
                 pdn_position_move: Optional[str] = None, steps_move: Optional[List[int]] = None,
                 li_api_move: Optional[List[str]] = None, li_one_move: Optional[str] = None,
                 has_captures: Optional[bool] = None, possible_moves: Optional[List[List[List[int]]]] = None,
                 possible_captures: Optional[List[List[Optional[int]]]] = None, hub_to_pdn_pseudolegal: bool = False,
                 variant: Optional[str] = None, is_null: Optional[bool] = None) -> None:
        self.ambiguous: Optional[bool] = None
        self.captures = None
        self.to_algebraic_variant = None
        self.has_captures = has_captures
        if (possible_moves is None or possible_captures is None) and board:
            self.possible_moves, self.possible_captures = board._legal_moves_board()
        else:
            self.possible_moves = possible_moves
            self.possible_captures = possible_captures
        if board and board.variant in ['russian', 'brazilian', 'turkish']:
            self.to_algebraic_variant = board.variant
        self.hub_to_pdn_pseudolegal = hub_to_pdn_pseudolegal
        self.variant = variant
        if not self.variant and board:
            self.variant = board.variant
        self.original_pdn_move = pdn_move
        self.is_null = is_null

        if self.is_null is None:
            self.is_null = (board_move == [[0, 0]] or hub_move == '0-0' or hub_position_move == '0000' or
                            pdn_move == '0-0' or pdn_position_move == '0000' or steps_move == [0, 0] or
                            li_api_move == ['0000'] or li_one_move == '0000')

        if ((board_move or hub_move or hub_position_move or pdn_move or pdn_position_move or steps_move or
             li_api_move or li_one_move) and not self.is_null):
            if self.possible_moves:
                self.board_move = self._to_board(board_move, hub_move, hub_position_move, pdn_move, pdn_position_move,
                                                 steps_move, li_api_move, li_one_move)
                self.captures = self.possible_captures[self.possible_moves.index(self.board_move)]
                self.captures = [] if self.captures[0] is None else self.captures
                self.has_captures = bool(self.captures)
                self._from_board(hub_move, hub_position_move, pdn_move, pdn_position_move, steps_move, li_api_move,
                                 li_one_move)
            else:
                self._no_board(board_move, hub_move, hub_position_move, pdn_move, pdn_position_move, steps_move,
                               li_api_move, li_one_move)
        elif self.is_null:
            self.board_move = [[0, 0]]
            self.hub_move = '0-0'
            self.hub_position_move = '0000'
            self.pdn_move = '0-0'
            self.pdn_position_move = '0000'
            self.steps_move = [0, 0]
            self.li_api_move = ['0000']
            self.li_one_move = '0000'

    def _to_board(self, board_move_given: Optional[List[List[int]]] = None, hub_move: Optional[str] = None,
                  hub_position_move: Optional[str] = None, pdn_move: Optional[str] = None,
                  pdn_position_move: Optional[str] = None, steps_move: Optional[List[int]] = None,
                  li_api_move: Optional[List[str]] = None, li_one_move: Optional[str] = None) -> List[List[int]]:
        """Convert the move to a board_move. Requires a Board() object to make the conversions."""

        board_move = [] if board_move_given is None else board_move_given
        # Hub related move

        if hub_move:
            # Hub move
            if "-" in hub_move:
                hub_position_move = "".join(list(map(self._make_len_2, hub_move.split("-"))))
            else:
                hub_position_move = "".join(list(map(self._make_len_2, hub_move.split("x"))))

        # PDN related move

        if pdn_move:
            # PDN move
            pdn_move = _algebraic_to_number(pdn_move, variant=self.variant)

            if "-" in pdn_move:
                pdn_position_move = "".join(list(map(self._make_len_2, pdn_move.split("-"))))
            else:
                pdn_position_move = "".join(list(map(self._make_len_2, pdn_move.split("x"))))

        # Hub related move

        if hub_position_move:
            # Hub position move

            # Order the captures
            hub_position_move_to_use = hub_position_move[:4] + self._sort_captures(
                [int(hub_position_move[i:i + 2]) for i in range(4, len(hub_position_move), 2)])

            moves_li_board = {}
            for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(
                    possible_move[-1][1]) + self._sort_captures(possible_capture)
                moves_li_board[li_move] = possible_move
            board_move = moves_li_board[hub_position_move_to_use]

        # PDN related move

        elif pdn_position_move:
            # PDN position move
            moves_li_board = {}
            if len(pdn_position_move) == 4:
                self.ambiguous = False
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1])
                    moves_li_board[li_move] = possible_move
            else:
                self.ambiguous = True
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    steps = [possible_move[0][0]]
                    for move in possible_move:
                        steps.append(move[1])
                    li_move = "".join(list(map(self._make_len_2, steps)))
                    moves_li_board[li_move] = possible_move
            board_move = moves_li_board[pdn_position_move]

        # Board related moves

        elif steps_move:
            # steps_move
            board_move = []
            for index in range(1, len(steps_move)):
                board_move.append([steps_move[index - 1], steps_move[index]])

        elif li_api_move:
            # li_api_move
            board_move = []
            for move in li_api_move:
                board_move.append([int(move[:2]), int(move[2:])])

        elif li_one_move:
            # li_one_move
            steps = [int(li_one_move[i:i + 2]) for i in range(0, len(li_one_move), 2)]
            board_move = []
            for index in range(1, len(steps)):
                board_move.append([steps[index - 1], steps[index]])

        return board_move

    def _from_board(self, hub_move: Optional[str] = None, hub_position_move: Optional[str] = None,
                    pdn_move: Optional[str] = None, pdn_position_move: Optional[str] = None,
                    steps_move: Optional[List[int]] = None, li_api_move: Optional[List[str]] = None,
                    li_one_move: Optional[str] = None) -> None:
        """Convert the move to all other move types. Requires a Board() object to make the conversions."""

        hub_move = "" if hub_move is None else hub_move
        hub_position_move = "" if hub_position_move is None else hub_position_move
        pdn_move = "" if pdn_move is None else pdn_move
        pdn_position_move = "" if pdn_position_move is None else pdn_position_move
        steps_move = [] if steps_move is None else steps_move
        li_api_move = [] if li_api_move is None else li_api_move
        li_one_move = "" if li_one_move is None else li_one_move

        # Board related move

        if not steps_move:
            # steps_move
            steps = [self.board_move[0][0]]
            for move in self.board_move:
                steps.append(move[1])
            steps_move = steps

        # Hub related moves

        if not hub_position_move:
            # Hub position move
            hub_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])
            if self.captures:
                sorted_captures = self._sort_captures(self.captures)
                sorted_captures_list = [sorted_captures[i:i + 2] for i in range(0, len(sorted_captures), 2)]
                hub_position_move += "".join(list(map(self._make_len_2, sorted_captures_list)))

        if not hub_move:
            # Hub move
            positions = [hub_position_move[i:i + 2] for i in range(0, len(hub_position_move), 2)]
            separator = "x" if self.captures else "-"
            hub_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        # PDN related moves

        if not pdn_position_move:
            # starts_endings contains the start position and the end position of every possible move.
            # It is used to detect ambiguous PDN moves, so we know when we have to expand the PDN move.
            # e.g. Use 3x14x23 instead of 3x23.

            # PDN position move

            start_end = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])
            same_start_ends = 0
            moves_with_same_start_end_and_captures = []
            for possible_move, possible_captures in zip(self.possible_moves, self.possible_captures):
                possible_start_end = self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1])
                if possible_start_end == start_end:
                    same_start_ends += 1
                    possible_steps_move = [possible_move[0][0]]
                    for semi_move in possible_move:
                        possible_steps_move.append(semi_move[1])
                    if possible_captures == self.captures:
                        moves_with_same_start_end_and_captures.append(possible_steps_move)
            self.ambiguous = same_start_ends >= 2
            if self.ambiguous:
                # From FMJD (https://pdn.fmjd.org/grammar.html#pdn-3-0-grammar):
                # Disambiguated capture sequences have to specify a complete sequence of intermediate squares
                # along the path of the capture. If there is a change in direction, an intermediate square is
                # the square where a turn in direction was made. If there was not a change in direction,
                # the intermediate square is the square immediately behind a captured piece.
                #
                # This means that if we continue going in the same direction,
                # the intermediate square is the square immediately behind the captured piece.
                # pydraughts supports providing a square not immediately behind the captured piece, but when we
                # convert it to a PDN move, we have to change it to the square immediately behind the captured piece.
                correct_start_ends = []
                # If it is equal to 1 then there are other moves with the same start and end, but not the same captures.
                # This can happen in russian draughts.
                if len(moves_with_same_start_end_and_captures) > 1:
                    for index in range(1, len(steps_move) - 1):
                        closest_distance = 1000

                        for possible_steps_move in moves_with_same_start_end_and_captures:
                            distance = abs(possible_steps_move[index - 1] - possible_steps_move[index])

                            # New closest square
                            if distance < closest_distance:
                                closest_distance = distance
                                correct_start_ends = [possible_steps_move]
                            # The same intermediate square.
                            # This means that the difference is in a later intermediate square.
                            elif distance == closest_distance:
                                correct_start_ends.append(possible_steps_move)
                            moves_with_same_start_end_and_captures = correct_start_ends

                    correct_move = correct_start_ends[0]
                else:
                    correct_move = steps_move

                pdn_position_move = "".join(list(map(self._make_len_2, correct_move)))
            else:
                pdn_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])

        if not pdn_move:
            # PDN move
            positions = [pdn_position_move[i:i + 2] for i in range(0, len(pdn_position_move), 2)]
            separator = "x" if self.captures else "-"
            pdn_move = separator.join(list(map(lambda position: str(int(position)), positions)))
            if self.to_algebraic_variant:
                pdn_move = _number_to_algebraic(pdn_move, variant=self.variant)

        # Board related moves

        if not li_api_move:
            # li_api_move
            li_api_move = []
            for move in self.board_move:
                li_api_move.append(self._make_len_2(move[0]) + self._make_len_2(move[1]))

        if not li_one_move:
            # li_one_move
            li_one_move = "".join(list(map(self._make_len_2, steps_move)))

        self.hub_move = hub_move
        self.hub_position_move = hub_position_move
        self.pdn_move = pdn_move
        self.pdn_position_move = pdn_position_move
        self.steps_move = steps_move
        self.li_api_move = li_api_move
        self.li_one_move = li_one_move


class Board:
    """A draughts game which considers the variant."""
    def __init__(self, variant: str = "standard", fen: str = "startpos"):
        self.variant = _convert_variant_names(variant)
        self._game = Game(variant, fen_from_variant(fen, variant) if fen != "startpos" else fen)
        self.initial_fen = fen_to_variant(self._game.initial_fen, self.variant)
        self.move_stack: List[Move] = []
        self.fens = [self.initial_fen]

        self._last_non_reversible_fen = self.initial_fen
        self._last_non_reversible_fens: List[str] = [self._last_non_reversible_fen]
        self._reversible_moves: List[Move] = []

    def copy(self) -> Board:
        """Copy the board (transfers all data)."""
        # At least 6 times faster than deepcopy.
        new_board: Board = pickle.loads(pickle.dumps(self, -1))
        return new_board

    def pop(self) -> Board:
        """Undo the last move."""
        self._game.pop()
        self.fens.pop()
        self.move_stack.pop()
        self._last_non_reversible_fens.pop()
        self._last_non_reversible_fen = self._last_non_reversible_fens[-1]
        if self._reversible_moves:
            self._reversible_moves.pop()
        return self

    def push(self, move: Move) -> Board:
        """Make a move."""
        self.move_stack.append(move)
        board_move = move.board_move.copy()
        for index, steps in enumerate(board_move):
            board_move[index] = list(map(lambda square: int(move_from_variant(str(square), variant=self.variant)), steps))
        self._game.push(board_move)
        self.fens.append(fen_to_variant(self._game.get_li_fen(), self.variant))
        if self._game.reversible_moves:
            self._reversible_moves.append(move)
        else:
            self._reversible_moves = []
            self._last_non_reversible_fen = self.fen
        self._last_non_reversible_fens.append(self._last_non_reversible_fen)
        return self

    def null(self) -> Board:
        """Play a null move."""
        self.push(Move(self, steps_move=[0, 0]))
        self.fens.append(fen_to_variant(self._game.get_li_fen(), self.variant))
        return self

    def winner(self) -> Optional[int]:
        """
        Get the player who won.
        :returns: WHITE if white won, BLACK if black won, 0 if it is a draw, and None if the game hasn't ended.
        """
        winner = self._game.get_winner()
        return 3 - winner if self.variant == "english" and winner else winner

    def is_over(self) -> bool:
        """Get if the game is over."""
        return self.winner() is not None

    @property
    def fen(self) -> str:
        """Get the fen of the current position."""
        return self.fens[-1]

    def _legal_moves_board(self) -> Tuple[List[List[List[int]]], List[List[Optional[int]]]]:
        """Get the legal moves for the current position in board_move format."""
        legal_moves, legal_captures = self._game.legal_moves()
        for move_index, move_capture in enumerate(zip(legal_moves, legal_captures)):
            for index, steps in enumerate(move_capture[0]):
                legal_moves[move_index][index] = list(map(lambda square: int(move_to_variant(
                    str(square), variant=self.variant, to_algebraic=False)), steps))
            legal_captures[move_index] = list(map(lambda square: None if square is None else int(move_to_variant(
                str(square), variant=self.variant, to_algebraic=False)), move_capture[1]))
        return legal_moves, legal_captures

    def legal_moves(self) -> List[Move]:
        """Get the legal moves for the current position."""
        legal_board_moves, legal_captures = self._legal_moves_board()
        legal_moves: List[Move] = list(map(lambda board_move: Move(
            board_move=board_move, possible_moves=legal_board_moves, possible_captures=legal_captures), legal_board_moves))
        return legal_moves

    @property
    def turn(self) -> int:
        """Get whose turn it is."""
        game_turn = self._game.whose_turn()
        return 3 - game_turn if self.variant == "english" else game_turn

    def __repr__(self) -> str:
        """Get a visual representation of the board."""
        game_repr = self._game.__repr__()
        if self.variant == "english":
            # game_repr = game_repr[::-1]
            game_repr = game_repr.replace("b", "z").replace("B", "Z")
            game_repr = game_repr.replace("w", "b").replace("W", "B")
            game_repr = game_repr.replace("z", "w").replace("Z", "W")
        return game_repr



import draughts.engines.dxp_communication.dxp_communication as dxp_communication
import draughts
import draughts.engine
import subprocess
import os
import signal
import threading
import time
import logging
from typing import Optional, Dict, Union, List, Any, Tuple

logger = logging.getLogger("pydraughts")


class DXPEngine:
    def __init__(self, command: Union[List[str], str, None] = None,
                 options: Optional[Dict[str, Union[str, int, bool]]] = None, initial_time: int = 0,
                 cwd: Optional[str] = None, ENGINE: int = 5) -> None:
        if options is None:
            options = {}
        self.initial_time = initial_time
        self.max_moves = 0
        self.ip = '127.0.0.1'
        self.port = '27531'
        self.command = command
        self.engine_opened = True  # Whether the engine is already open or pydraughts should open it
        self.wait_to_open_time = 10
        self.ENGINE = ENGINE
        self.info: Dict[str, Any] = {}
        self.id: Dict[str, str] = {}
        self.sender = dxp_communication.Sender()
        self.receiver = self.sender.receiver
        self.game_started = False
        self.exit = False

        self.configure(options)

        if not self.engine_opened:
            cwd = cwd or os.getcwd()
            cwd = os.path.realpath(os.path.expanduser(cwd))
            if type(command) == str:
                command = [command]
            command = list(filter(None, command))
            command[0] = os.path.realpath(os.path.expanduser(command[0]))
            command[0] = '"' + command[0] + '"'
            command = ' '.join(command)
            self.command = command
            self.p = self._open_process(command, cwd)
            self.engine_receive_thread = threading.Thread(target=self._recv)
            self.engine_receive_thread.start()

        self.start_time = time.perf_counter_ns()
        self.quit_time = 0

    def setoption(self, name: str, value: Union[str, int, bool]) -> None:
        """Set a DXP option."""
        if name == 'engine-opened':
            self.engine_opened = bool(value)
        elif name == 'ip':
            self.ip = str(value)
        elif name == 'port':
            self.port = str(value)
        elif name == 'wait-to-open-time':
            self.wait_to_open_time = int(value)
        elif name == 'max-moves':
            self.max_moves = int(value)
        elif name == 'initial-time':
            self.initial_time = int(value)

    def configure(self, options: Dict[str, Union[str, int, bool]]) -> None:
        """Configure many options at once."""
        for name, value in options.items():
            self.setoption(name, value)

    def _open_process(self, command: str, cwd: Optional[str] = None, shell: bool = True,
                      _popen_lock: Any = threading.Lock()) -> subprocess.Popen:
        """Open the engine process."""
        kwargs = {
            "shell": shell,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "stdin": subprocess.PIPE,
            "bufsize": 1,  # Line buffered
            "universal_newlines": True,
        }
        logger.debug(f'command: {command}, cwd: {cwd}')

        if cwd is not None:
            kwargs["cwd"] = cwd

        # Prevent signal propagation from parent process
        try:
            # Windows
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
        except AttributeError:
            # Unix
            kwargs["preexec_fn"] = os.setpgrp

        with _popen_lock:  # Work around Python 2 Popen race condition
            return subprocess.Popen(command, **kwargs)

    def kill_process(self) -> None:
        """Kill the engine process."""
        if not self.engine_opened:
            wait_time = self.quit_time / 1e9 + 10 - time.perf_counter_ns() / 1e9
            logger.debug(f'wait time before killing: {wait_time}')
            if wait_time > 0:
                time.sleep(wait_time)
            self.exit = True
            try:
                # Windows
                logger.debug("Killing Windows.")
                self.p.send_signal(signal.CTRL_BREAK_EVENT)
            except AttributeError:
                # Unix
                logger.debug("Killing UNIX.")
                os.killpg(self.p.pid, signal.SIGTERM)
                time.sleep(7)
                os.killpg(self.p.pid, signal.SIGKILL)

            self.p.communicate()
            self.engine_receive_thread.join()

    def _connect(self) -> None:
        """Connect to the engine."""
        if not self.engine_opened:
            wait_time = self.start_time / 1e9 + self.wait_to_open_time - time.perf_counter_ns() / 1e9
            logger.debug(f'wait time before connecting: {wait_time}')
            if wait_time > 0:
                time.sleep(wait_time)
        self.sender.connect(self.ip, int(self.port))

    def _start(self, board: draughts.Board, game_time: int) -> None:
        """Start the game."""
        self._connect()
        engine_color = 'B' if board.turn == draughts.WHITE else 'W'
        game_time = int(round(game_time // 60))
        moves = self.max_moves
        self.sender.setup(board.initial_fen, board.variant)
        self.sender.gamereq(engine_color, game_time, moves)
        accepted = self._recv_accept()
        logger.debug(f'Aceepted: {accepted}')
        self.id["name"] = self.sender.current.engineName

    def _recv(self) -> None:
        """Receive a line from the engine, if the engine is opened by pydraughts."""
        # The engine doesn't work otherwise.
        while True:
            try:
                line = self.p.stdout.readline()

                line = line.rstrip()

                if line:
                    logger.debug(f"{self.ENGINE} %s >> %s {self.p.pid} {line}")
            except ValueError as err:
                if self.exit:
                    break
                else:
                    raise err

    def _recv_accept(self) -> bool:
        """Get if the game was accepted."""
        while True:
            if self.receiver.accepted is not None:
                return self.receiver.accepted

    def _recv_move(self) -> Optional[draughts.Move]:
        """Receive the engine move."""
        while True:
            if not self.receiver.listening:
                break
            if self.receiver.last_move_changed:
                logger.debug(f'new last move: {self.receiver.last_move.board_move}')
                return self.receiver.last_move
        return None

    def _recv_backreq(self) -> bool:
        """Get if the backreq was accepted."""
        while True:
            if self.receiver.backreq_accepted is not None:
                return self.receiver.backreq_accepted

    def takeback(self, move: int, color: int) -> Tuple[bool, draughts.Board, Any]:
        """
        Attempt to take back moves.
        Warning: If it is the engine's turn to play in the new position, it will attempt to move.
        """
        self.receiver.backreq_accepted = None
        ply = (move - 1) * 2 + (0 if color == draughts.WHITE else 1)
        remove_count = 0
        best_move = None
        moves = list(map(lambda old_move: old_move.steps_move, self.sender.current.pos.move_stack))
        logger.debug(f"Move stack before removing: {moves}")
        self.sender.backreq(move, color)
        backreq = self._recv_backreq()
        if backreq:
            while len(self.sender.current.pos.move_stack) > ply:
                remove_count += 1
                self.sender.current.pos.pop()
                moves = list(map(lambda old_move: old_move.steps_move, self.sender.current.pos.move_stack))
                logger.debug(f"Move stack after removing {remove_count} moves: {moves}")
            self.receiver.takeback_in_progress = False
            new_board = self.sender.current.pos.copy()
            if self.sender.current.engine_color == self.sender.current.get_color():  # It is the engine's turn to play.
                best_move = self._recv_move()
        else:
            new_board = self.sender.current.pos.copy()
        return backreq, new_board, draughts.engine.PlayResult(best_move, None, {})

    def play(self, board: draughts.Board) -> Any:
        """Engine search."""
        if not self.game_started:
            self._start(board, self.initial_time)
            self.game_started = True
        if board.move_stack:
            self.sender.send_move(board.move_stack[-1])
        best_move = self._recv_move()
        return draughts.engine.PlayResult(best_move, None, {})

    def quit(self) -> None:
        """Quit the engine."""
        self.sender.gameend()
        self.sender.disconnect()
        self.quit_time = time.perf_counter_ns()



from __future__ import annotations
from functools import reduce
from typing import List, Dict, Any
from draughts.core.piece import Piece

WHITE = 2
BLACK = 1


class BoardSearcher:

    def build(self, board: Any) -> None:
        """Build the searcher."""
        self.board = board
        self.uncaptured_pieces = list(filter(lambda piece: not piece.captured, board.pieces))
        self.open_positions: List[int] = []
        self.filled_positions: List[int] = []
        self.player_positions: Dict[int, List[int]] = {}
        self.player_pieces: Dict[int, List[Piece]] = {}
        self.position_pieces: Dict[int, Piece] = {}

        self.build_filled_positions()
        self.build_open_positions()
        self.build_player_positions()
        self.build_player_pieces()
        self.build_position_pieces()

    def build_filled_positions(self) -> None:
        """Find the filled positions (squares which have a piece)."""
        self.filled_positions = reduce((lambda open_positions, piece: open_positions + [piece.position]),
                                       self.uncaptured_pieces, [])

    def build_open_positions(self) -> None:
        """Find the open positions (empty squares)."""
        self.open_positions = list(set(range(1, self.board.position_count + 1)).difference(self.filled_positions))

    def build_player_positions(self) -> None:
        """Find the positions where each player has a piece."""
        self.player_positions = {
            1: reduce((lambda positions, piece: positions + ([piece.position] if piece.player == BLACK else [])),
                      self.uncaptured_pieces, []),
            2: reduce((lambda positions, piece: positions + ([piece.position] if piece.player == WHITE else [])),
                      self.uncaptured_pieces, [])
        }

    def build_player_pieces(self) -> None:
        """Find all the pieces of both players."""
        self.player_pieces = {
            BLACK: reduce((lambda pieces, piece: pieces + ([piece] if piece.player == BLACK else [])),
                          self.uncaptured_pieces, []),
            WHITE: reduce((lambda pieces, piece: pieces + ([piece] if piece.player == WHITE else [])),
                          self.uncaptured_pieces, [])
        }

    def build_position_pieces(self) -> None:
        """Make a dict where the key is the square and the value is the piece in this square."""
        self.position_pieces = {piece.position: piece for piece in self.uncaptured_pieces}

    def get_pieces_by_player(self, player_number: int) -> List[Piece]:
        """Get all the pieces of one player."""
        return self.player_pieces[player_number]

    def get_positions_by_player(self, player_number: int) -> List[int]:
        """Get the positions of one player's pieces."""
        return self.player_positions[player_number]

    def get_pieces_in_play(self) -> List[Piece]:
        """
        Get pieces in play. They are: All the pieces of the player playing now except when a piece is
        in the middle of a multi-capture, so it has already captured one or more pieces, but it can capture more,
        where we only return the piece that is in the middle of the multi-capture.
        """
        return (self.player_pieces[self.board.player_turn] if not self.board.piece_requiring_further_capture_moves else
                [self.board.piece_requiring_further_capture_moves])

    def get_piece_by_position(self, position: int) -> Piece:
        """Get the piece given its position."""
        return self.position_pieces[position]



from draughts.core.variant import Board, Move, WHITE, BLACK

__all__ = ["Board", "Move", "WHITE", "BLACK"]

__author__ = "Ioannis Pantidis"
__copyright__ = "2021-2024, " + __author__
__version__ = "0.6.6"



import os
from typing import Optional, Union, Tuple, Dict, Any
import draughts

from msl.loadlib import Client64


class Engine32(Client64):
    """64 bit client to communicate with the 32 bit server that is running the old Checkerboard engine."""
    def __init__(self, command: str) -> None:

        command = command.replace('\\', '\\\\')

        super(Engine32, self).__init__(module32='engine_server', append_sys_path=os.path.dirname(__file__), timeout=15.,
                                       dll_name=command)

    def enginecommand(self, command: str) -> Tuple[bytes, int]:
        """Send an enginecommand to the engine."""
        response: Tuple[bytes, int] = self.request32('enginecommand', command)
        return response

    def getmove(self, game: draughts.Board, maxtime: Union[int, float, None] = None, time: Union[int, float, None] = None,
                increment: Union[int, float, None] = None, movetime: Union[int, float, None] = None
                ) -> Tuple[Optional[str], bytes, Dict[str, Any], int]:
        """Send a getmove to the engine."""
        assert maxtime is not None or time is not None or movetime is not None
        response: Tuple[Optional[str], bytes, Dict[str, Any], int] = self.request32('getmove', game, maxtime, time,
                                                                                    increment, movetime)
        return response



from draughts.engine import HubEngine, DXPEngine, CheckerBoardEngine, Limit
from draughts import Board, WHITE, BLACK
from draughts.PDN import PDNWriter
from typing import List, Tuple, Dict, Any, Union, Optional
import datetime
import itertools
import time
import logging

logger = logging.getLogger("pydraughts")


class RoundRobin:
    def __init__(self, filename: str, players: List[Tuple[Union[str, List[str]], str, Dict[str, Any], Optional[str]]],
                 start_time: Union[int, float], increment: Union[int, float] = 0, variant: str = "standard",
                 games_per_pair: int = 2, starting_fen: str = "startpos", max_moves: int = 300) -> None:
        self.filename = filename
        self.players = players
        self.start_time = start_time
        self.increment = increment
        self.variant = variant
        self.games_per_pair = games_per_pair
        self.starting_fen = starting_fen
        self.max_moves = max_moves
        self.player_count = len(self.players)
        self.int_players = list(range(self.player_count))
        self.results = [[0, 0, 0] for _ in range(self.player_count)]
        self.scores = [0] * self.player_count
        self.complete_results: List[List[Tuple[Tuple[int, int, int], Tuple[int, int, int]]]] = []
        self.pairs: List[Tuple[int, int]] = list(itertools.combinations(self.int_players, 2))
        logger.debug(f"There are {len(self.pairs)} possible pairs.")
        self.complete_pairs: List[List[Tuple[int, int]]] = []
        self.get_complete_pairs()
        logger.debug(f"There will be {len(self.complete_pairs)} games.")
        date = datetime.datetime.now()
        year = str(date.year).zfill(4)
        month = str(date.month).zfill(2)
        day = str(date.day).zfill(2)
        self.tags = {"Event": "Tournament", "Site": "pydraughts", "EventDate": f"{year}.{month}.{day}",
                     "ResultType": "International", "WhiteType": "program", "BlackType": "program", "GameType": "20",
                     "TimeControl": f"{self.start_time}+{self.increment}"}
        if self.starting_fen != "startpos":
            self.tags["FEN"] = self.starting_fen
        self.round = 0

        self.latest_round_results: List[Tuple[Tuple[int, int, int], Tuple[int, int, int]]] = []
        self.latest_complete_results: List[List[Tuple[Tuple[int, int, int], Tuple[int, int, int]]]] = []

    def get_complete_pairs(self) -> None:
        pairs = self.pairs.copy()
        for match in range(self.games_per_pair):
            self.complete_pairs.append(pairs.copy())
            pairs = list(map(lambda _pair: (_pair[1], _pair[0]), pairs))

    def get_engine(self, command: Union[str, List[str]], protocol: str, options: Dict[str, Any], cwd: Optional[str],
                   ) -> Union[HubEngine, DXPEngine, CheckerBoardEngine]:
        engine: Union[HubEngine, DXPEngine, CheckerBoardEngine]
        if protocol.lower() == "hub":
            engine = HubEngine(command, cwd=cwd)
            engine.configure(options)
        elif protocol.lower() == "dxp":
            options["initial-time"] = self.start_time
            options["max-moves"] = self.max_moves
            engine = DXPEngine(command, options, cwd=cwd)
        elif protocol.lower() == "cb" or protocol.lower() == "checkerboard":
            engine = CheckerBoardEngine(command)
            engine.configure(options)
        else:
            raise ValueError(f"There is no protocol `{protocol}`.")
        return engine

    def play(self) -> List[int]:
        while self.round < self.games_per_pair:
            logger.debug(f"Playing round {self.round + 1}/{self.games_per_pair}")
            self.play_round()
            self.round += 1
        for player in range(self.player_count):
            self.scores[player] = self.results[player][0] * 2 + self.results[player][1]
        return self.scores

    def play_round(self) -> None:
        round_results = []
        for game_number, match in enumerate(self.complete_pairs[self.round]):
            logger.debug(f"Playing game {game_number+1}/{len(self.complete_pairs[self.round])} in {self.round+1}th round.")
            result1, result2 = self.play_game(self.players[match[0]], self.players[match[1]], game_number+1)
            round_results.append((result1, result2))
            # Player 1
            self.results[match[0]][0] += result1[0]
            self.results[match[0]][1] += result1[1]
            self.results[match[0]][2] += result1[2]

            # Player 2
            self.results[match[1]][0] += result2[0]
            self.results[match[1]][1] += result2[1]
            self.results[match[1]][2] += result2[2]
            self.latest_round_results = round_results.copy()
            logger.debug(f"Round results until now: {self.latest_round_results}")
        self.complete_results.append(round_results)
        self.latest_complete_results = self.complete_results.copy()
        logger.debug(f"Complete results until now: {self.latest_complete_results}")

    def play_game(self, player_1_info: Tuple[Union[str, List[str]], str, Dict[str, Any]],
                  player_2_info: Tuple[Union[str, List[str]], str, Dict[str, Any]], game_number: int
                  ) -> Tuple[Tuple[int, int, int], Tuple[int, int, int]]:
        logger.debug(f"Player 1: '{player_1_info[0]}', Player 2: '{player_2_info[0]}'")
        board = Board(self.variant, self.starting_fen)
        player_1 = self.get_engine(*player_1_info)
        player_2 = self.get_engine(*player_2_info)
        if isinstance(player_1, HubEngine):
            player_1.init()
        if isinstance(player_2, HubEngine):
            player_2.init()
        player_1_limit = Limit(self.start_time, self.increment)
        player_2_limit = Limit(self.start_time, self.increment)
        max_moves = self.max_moves
        if max_moves == 0:
            max_moves = 10000
        while not board.is_over() and len(board.move_stack) < max_moves:
            logger.info(f'move: {len(board.move_stack)}')
            if board.turn == WHITE:
                start = time.perf_counter_ns()
                if isinstance(player_1, HubEngine):
                    best_move = player_1.play(board, player_1_limit, False)
                elif isinstance(player_1, DXPEngine):
                    best_move = player_1.play(board)
                else:  # Checkerboard
                    best_move = player_1.play(board, player_1_limit)
                end = time.perf_counter_ns()
                player_1_limit.time = player_1_limit.time - (end - start) / 1e9 + player_1_limit.inc
            else:
                start = time.perf_counter_ns()
                if isinstance(player_2, HubEngine):
                    best_move = player_2.play(board, player_2_limit, False)
                elif isinstance(player_2, DXPEngine):
                    best_move = player_2.play(board)
                else:  # Checkerboard
                    best_move = player_2.play(board, player_2_limit)
                end = time.perf_counter_ns()
                player_2_limit.time = player_2_limit.time - (end - start) / 1e9 + player_2_limit.inc
            if best_move.move:
                board.push(best_move.move)
            else:
                break
        if not isinstance(player_1, CheckerBoardEngine):
            player_1.quit()
        if not isinstance(player_2, CheckerBoardEngine):
            player_2.quit()
        player_1.kill_process()
        player_2.kill_process()
        winner = board.winner()
        game_ending = "1-1"
        if winner == WHITE:
            game_ending = "2-0"
        elif winner == BLACK:
            game_ending = "0-2"

        # Write PDN
        tags = self.tags.copy()
        tags["Result"] = game_ending
        tags["Round"] = f"{self.round + 1}.{game_number}"
        tags["White"] = str(player_1_info[0])
        tags["Black"] = str(player_2_info[0])
        tags["PlyCount"] = str(len(board.move_stack))
        date_tag, time_tag, utc_date_tag, utc_time_tag = self._get_date_tags()
        tags["Date"] = date_tag
        tags["Time"] = time_tag
        tags["UTCDate"] = utc_date_tag
        tags["UTCTime"] = utc_time_tag
        PDNWriter(self.filename, board, tags=tags, game_ending=game_ending)

        # Return result
        if winner == WHITE:  # Player 1 won
            logger.debug("Player 1 won")
            return (1, 0, 0), (0, 0, 1)
        elif winner == BLACK:  # player 2 won
            logger.debug("Player 2 won")
            return (0, 0, 1), (1, 0, 0)
        else:  # Draw
            logger.debug("Game ended in a draw.")
            return (0, 1, 0), (0, 1, 0)

    def _get_date_tags(self) -> Tuple[str, str, str, str]:
        date = datetime.datetime.now()
        year = str(date.year).zfill(4)
        month = str(date.month).zfill(2)
        day = str(date.day).zfill(2)
        hour = str(date.hour).zfill(2)
        minute = str(date.minute).zfill(2)
        second = str(date.second).zfill(2)
        date_tag = f"{year}.{month}.{day}"
        time_tag = f"{hour}:{minute}:{second}"
        date = datetime.datetime.utcnow()
        year = str(date.year).zfill(4)
        month = str(date.month).zfill(2)
        day = str(date.day).zfill(2)
        hour = str(date.hour).zfill(2)
        minute = str(date.minute).zfill(2)
        second = str(date.second).zfill(2)
        utc_date_tag = f"{year}.{month}.{day}"
        utc_time_tag = f"{hour}:{minute}:{second}"
        return date_tag, time_tag, utc_date_tag, utc_time_tag

    def get_standings(self) -> List[Tuple[int, int]]:
        standings = list(zip(self.scores, self.int_players))
        standings.sort(reverse=True)
        return standings

    def print_standings(self) -> None:
        standings = self.get_standings()
        for place, engine in enumerate(standings):
            logger.debug(f"{place + 1}th place: {self.players[engine[1]][0]} with {engine[0]} points.")
            print(f"{place+1}th place: {self.players[engine[1]][0]} with {engine[0]} points.")
